/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2016 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.layout.auto;

import static org.fortiss.tooling.base.utils.LayoutDataUtils.getNodeSize;
import static org.fortiss.tooling.base.utils.LayoutDataUtils.setNodePosition;

import java.util.Map.Entry;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.BasicEMap;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.EMap;
import org.fortiss.tooling.base.model.base.EntryConnectorBase;
import org.fortiss.tooling.base.model.base.ExitConnectorBase;
import org.fortiss.tooling.base.model.element.IConnection;
import org.fortiss.tooling.base.model.element.IConnector;
import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.base.model.layout.Dimension;
import org.fortiss.tooling.base.model.layout.EOrientation;
import org.fortiss.tooling.base.model.layout.ILayoutData;
import org.fortiss.tooling.base.model.layout.ILayoutedModelElement;
import org.fortiss.tooling.base.model.layout.Points;
import org.fortiss.tooling.base.model.layout.impl.PointsImpl;
import org.fortiss.tooling.base.ui.annotation.view.AnnotationViewPartBase;
import org.fortiss.tooling.base.ui.utils.LayoutDataUIUtils;
import org.fortiss.tooling.base.utils.LayoutDataUtils;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.service.IPersistencyService;

import de.cau.cs.kieler.core.alg.BasicProgressMonitor;
import de.cau.cs.kieler.core.alg.IKielerProgressMonitor;
import de.cau.cs.kieler.core.kgraph.KEdge;
import de.cau.cs.kieler.core.kgraph.KNode;
import de.cau.cs.kieler.core.kgraph.KPort;
import de.cau.cs.kieler.kiml.AbstractLayoutProvider;
import de.cau.cs.kieler.kiml.klayoutdata.KEdgeLayout;
import de.cau.cs.kieler.kiml.klayoutdata.KPoint;
import de.cau.cs.kieler.kiml.klayoutdata.KShapeLayout;
import de.cau.cs.kieler.kiml.options.Direction;
import de.cau.cs.kieler.kiml.options.LayoutOptions;
import de.cau.cs.kieler.kiml.options.PortSide;
import de.cau.cs.kieler.kiml.util.KimlUtil;
import de.cau.cs.kieler.klay.layered.LayeredLayoutProvider;

/**
 * <p>
 * {@link IAutoLayouter} based on <a
 * href="https://rtsys.informatik.uni-kiel.de/confluence/display/KIELER/Home">KIELER framework</a>.
 * </p>
 * 
 * <
 * 
 * Approach:
 * <ul>
 * <li>Generate a KIELER graph from the input model</li>
 * <li>Apply the KIELER <i>layered layout algorithm</i></li>
 * <li>Transfer generated layout to input model
 * </ul>
 * 
 * @author offtermatt, barner
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: 16F143B594364B7086DDC34B46E25EA6
 */
public class KielerAutoLayouter implements IAutoLayouter {

	/** Minimum mode width for layout. */
	private static final int NODE_MIN_WIDTH = 150;

	/** Height per port (used to derive minimum height of layouted element. */
	private static final int PORT_HEIGHT = 15;

	/** Mapping from {@link IHierarchicElement}s to {@link KNode}s. */
	private EMap<IHierarchicElement, KNode> modelElementsToKNodes;

	/** Mapping from outbound {@link IConnector}s to {@link KPort}s. */
	private EMap<IConnector, KPort> outboundConnectorsToKPorts;

	/** Mapping from inbound {@link IConnector}s to {@link KPort}s. */
	private EMap<IConnector, KPort> inboundConnectorsToKPorts;

	/** Used to map undirected {@link IConnector}s to {@link KPort}s */
	private EMap<IConnector, KPort> undirectedConnectorsToKPorts;

	/** Mapping from {@link IConnection}s to {@link KEdge}s. */
	private EMap<IConnection, KEdge> connectionsToKEdges;

	/** Mapping from global inbound {@link IConnector}s to {@link KPort}s. */
	private EMap<IConnector, KPort> globalInboundConnectorsToKPorts;

	/** Mapping from global outbound {@link IConnector}s to {@link KPort}s. */
	private EMap<IConnector, KPort> globalOutboundConnectorsToKPorts;

	/** {@inheritDoc} */
	@Override
	public void performAutoLayout(IHierarchicElement hierarchicElement) {

		boolean avUpdateEnabled = AnnotationViewPartBase.isUpdateEnabled();
		AnnotationViewPartBase.setUpdateEnabled(false);
		KNode parentNode = createKIELERGraph(hierarchicElement);

		KShapeLayout par = parentNode.getData(KShapeLayout.class);
		par.setProperty(LayoutOptions.DIRECTION, Direction.RIGHT);
		par.setProperty(LayoutOptions.INTERACTIVE, true);

		AbstractLayoutProvider layoutProvider = new LayeredLayoutProvider();
		IKielerProgressMonitor progressMonitor = new BasicProgressMonitor();
		layoutProvider.doLayout(parentNode, progressMonitor);

		applyLayout(parentNode, hierarchicElement);

		AnnotationViewPartBase.setUpdateEnabled(avUpdateEnabled);
	}

	/**
	 * Applies the layout defined in the {@link KNode} follows to given {@link IHierarchicElement}.
	 * 
	 * @param template
	 *            The {@link KNode} that stores the layout.
	 * @param hierarchicElement
	 *            The {@link IHierarchicElement} to which the layout should be applied.
	 */
	private void applyLayout(KNode template, IHierarchicElement hierarchicElement) {
		ITopLevelElement topLevelElement =
				IPersistencyService.getInstance().getTopLevelElementFor(hierarchicElement);
		topLevelElement.runAsCommand(() -> {
			filterBendPoints(hierarchicElement);

			for(IHierarchicElement curElement : hierarchicElement.getContainedElements()) {
				KNode curNode = modelElementsToKNodes.get(curElement);
				KShapeLayout curLayout = curNode.getData(KShapeLayout.class);
				LayoutDataUtils.setNodeLayoutData((ILayoutedModelElement)curElement,
						(int)curLayout.getXpos(), (int)curLayout.getYpos(),
						(int)curLayout.getWidth(), (int)curLayout.getHeight());
			}

			for(Entry<IConnection, KEdge> curEntry : connectionsToKEdges) {
				KEdge curEdge = curEntry.getValue();
				IConnection curConnection = curEntry.getKey();
				KEdgeLayout curLayout = curEdge.getData(KEdgeLayout.class);
				for(KPoint k : curLayout.getBendPoints()) {
					LayoutDataUIUtils.addBendPointToConnection(
							(ILayoutedModelElement)curConnection, (int)k.getX(), (int)k.getY());
				}
			}

			setPositionsForPorts(outboundConnectorsToKPorts);
			setPositionsForPorts(inboundConnectorsToKPorts);
			setPositionsForGlobalPorts(globalInboundConnectorsToKPorts);
			setPositionsForGlobalPorts(globalOutboundConnectorsToKPorts);
		});
	}

	/**
	 * Applies the layout of the specified {@link KPort}s to their matching (global)
	 * {@link IConnector}s. Global means in this case that the {@link IConnector}s do not have a
	 * corresponding node, i.e. because they contain connections to a higher-level model.
	 * 
	 * @param globalPorts
	 *            A bijection from {@link IConnector}s to {@link KPort}s
	 */
	private void setPositionsForGlobalPorts(EMap<IConnector, KPort> globalPorts) {
		for(Entry<IConnector, KPort> i : globalPorts) {
			IConnector port = i.getKey();
			KPort p = i.getValue();
			KNode temp = p.getNode();
			KShapeLayout lay = temp.getData(KShapeLayout.class);
			setNodePosition((ILayoutedModelElement)port, (int)lay.getXpos(), (int)lay.getYpos());
		}
	}

	/**
	 * Applies the layout of the specified {@link KPort}s to their matching {@link IConnector}s. The
	 * {@link IConnector}s must have corresponding nodes, i.e. they can not contain an
	 * {@link IConnection} to a higher-level model.
	 * 
	 * @param ports
	 *            A bijection from {@link IConnector}s to {@link KPort}s.
	 */
	private void setPositionsForPorts(EMap<IConnector, KPort> ports) {
		for(Entry<IConnector, KPort> ic : ports) {
			KPort kp = ic.getValue();
			IConnector port = ic.getKey();
			KShapeLayout portLayout = kp.getData(KShapeLayout.class);
			PortSide side = portLayout.getProperty(LayoutOptions.PORT_SIDE);
			EOrientation orientation = null;
			switch(side) {
				case NORTH:
					orientation = EOrientation.NORTH;
					break;
				case WEST:
					orientation = EOrientation.WEST;
					break;
				case EAST:
					orientation = EOrientation.EAST;
					break;
				case SOUTH:
					orientation = EOrientation.SOUTH;
					break;
			}
			LayoutDataUtils.setStickyConnectorLayoutData((ILayoutedModelElement)port, orientation,
					(int)portLayout.getYpos());
		}
	}

	/**
	 * Removes all bend-points in connections contained in the given {@link IHierarchicElement}.
	 * 
	 * @param hierarchicElement
	 */
	private void filterBendPoints(IHierarchicElement hierarchicElement) {
		EList<IConnection> clears = hierarchicElement.getConnections();
		for(IConnection c : clears) {
			EList<ILayoutData> l = ((ILayoutedModelElement)c).getLayoutData();
			for(ILayoutData i : l) {
				if(i instanceof Points || i instanceof PointsImpl) {
					((Points)i).getPoints().clear();
				}
			}
		}
	}

	/**
	 * Creates a KIELER {@link KNode} graph with the same structure as the given
	 * {@link IHierarchicElement}.
	 * 
	 * @param hierarchicElement
	 *            The {@link IHierarchicElement} for which the {@link KNode} graph should be created
	 * 
	 * @return A {@link KNode} graph with the same structure as the graph contained in the
	 *         {@link IHierarchicElement}.
	 */
	private KNode createKIELERGraph(IHierarchicElement hierarchicElement) {

		KNode parentNode = KimlUtil.createInitializedNode();

		modelElementsToKNodes = new BasicEMap<IHierarchicElement, KNode>();
		outboundConnectorsToKPorts = new BasicEMap<IConnector, KPort>();
		inboundConnectorsToKPorts = new BasicEMap<IConnector, KPort>();
		connectionsToKEdges = new BasicEMap<IConnection, KEdge>();
		undirectedConnectorsToKPorts = new BasicEMap<IConnector, KPort>();

		// create nodes
		for(IHierarchicElement curElement : hierarchicElement.getContainedElements()) {

			if(!(curElement instanceof ILayoutedModelElement)) {
				continue;
			}

			KNode curKNode = KimlUtil.createInitializedNode();

			KShapeLayout curLayout = curKNode.getData(KShapeLayout.class);

			EList<IConnector> inputPorts = new BasicEList<IConnector>();
			EList<IConnector> outputPorts = new BasicEList<IConnector>();
			EList<IConnector> genericPorts = new BasicEList<IConnector>();

			EList<IConnector> ports = curElement.getConnectors();
			for(IConnector curConnector : ports) {
				if(curConnector instanceof EntryConnectorBase) {
					inputPorts.add(curConnector);
				} else if(curConnector instanceof ExitConnectorBase) {
					outputPorts.add(curConnector);
				} else {
					outputPorts.add(curConnector);
					genericPorts.add(curConnector);
				}
			}

			Dimension dimension = getNodeSize((ILayoutedModelElement)curElement);

			int portNmbr = Math.max(inputPorts.size(), outputPorts.size());
			curLayout.setHeight(Math.max(portNmbr * PORT_HEIGHT, dimension.getHeight()));
			curLayout.setWidth(Math.max(NODE_MIN_WIDTH, dimension.getWidth()));

			curKNode.setParent(parentNode);
			modelElementsToKNodes.put(curElement, curKNode);
			for(IConnector curPort : outputPorts) {
				KPort k = KimlUtil.createInitializedPort();
				k.setNode(curKNode);
				outboundConnectorsToKPorts.put(curPort, k);
				if(genericPorts.contains(curPort)) {
					undirectedConnectorsToKPorts.put(curPort, k);
				}
			}
			for(IConnector curPort : inputPorts) {
				KPort k = KimlUtil.createInitializedPort();
				k.setNode(curKNode);
				inboundConnectorsToKPorts.put(curPort, k);
			}
		}

		// create virtual nodes for global ports
		globalInboundConnectorsToKPorts = new BasicEMap<IConnector, KPort>();
		globalOutboundConnectorsToKPorts = new BasicEMap<IConnector, KPort>();

		for(IConnector curPort : hierarchicElement.getConnectors()) {
			KNode virtual = KimlUtil.createInitializedNode();
			KShapeLayout virtualLayout = virtual.getData(KShapeLayout.class);
			virtualLayout.setHeight(50);
			virtualLayout.setWidth(50);
			virtual.setParent(parentNode);

			KPort k = KimlUtil.createInitializedPort();
			k.setNode(virtual);
			if(curPort instanceof ExitConnectorBase) {
				globalOutboundConnectorsToKPorts.put(curPort, k);
			} else {
				globalInboundConnectorsToKPorts.put(curPort, k);
				outboundConnectorsToKPorts.put(curPort, k);
			}
		}

		// create edges
		for(Entry<IConnector, KPort> out : outboundConnectorsToKPorts) {
			for(IConnection ic : out.getKey().getOutgoing()) {
				KEdge curEdge = KimlUtil.createInitializedEdge();
				KPort sourceport = out.getValue();

				curEdge.setSource(sourceport.getNode());
				curEdge.setSourcePort(sourceport);
				sourceport.getEdges().add(curEdge);
				KPort targetport = inboundConnectorsToKPorts.get(ic.getTarget());
				if(targetport == null) {
					targetport = globalOutboundConnectorsToKPorts.get(ic.getTarget());
				}
				if(targetport == null) {
					targetport = undirectedConnectorsToKPorts.get(ic.getTarget());
				}
				curEdge.setTargetPort(targetport);
				curEdge.setTarget(targetport.getNode());
				targetport.getEdges().add(curEdge);
				connectionsToKEdges.put(ic, curEdge);
			}
		}
		return parentNode;
	}
}
